# noqa: D100
from __future__ import annotations

from typing import TYPE_CHECKING

from sphinx.domains.python import PythonDomain

from pyproject_api import __version__

if TYPE_CHECKING:
    from docutils.nodes import Element
    from sphinx.application import Sphinx
    from sphinx.builders import Builder
    from sphinx.environment import BuildEnvironment
project = name = "pyproject_api"
company = "tox-dev"
project_copyright = f"{company}"
version, release = __version__, __version__.split("+")[0]

extensions = [
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.extlinks",
    "sphinx.ext.autodoc",
    "sphinx_autodoc_typehints",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
]
master_doc, source_suffix = "index", ".rst"

html_theme = "furo"
html_title, html_last_updated_fmt = "pyproject-api docs", "%Y-%m-%dT%H:%M:%S"
pygments_style, pygments_dark_style = "sphinx", "monokai"

autoclass_content, autodoc_typehints = "both", "none"
autodoc_default_options = {"members": True, "member-order": "bysource", "undoc-members": True, "show-inheritance": True}
inheritance_alias = {}

extlinks = {
    "issue": ("https://github.com/tox-dev/pyproject-api/issues/%s", "#%s"),
    "pull": ("https://github.com/tox-dev/pyproject-api/pull/%s", "PR #%s"),
    "user": ("https://github.com/%s", "@%s"),
}
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "packaging": ("https://packaging.pypa.io/en/latest", None),
}

nitpicky = True
nitpick_ignore = []


def setup(app: Sphinx) -> None:  # noqa: D103
    class PatchedPythonDomain(PythonDomain):
        def resolve_xref(  # noqa: PLR0913,PLR0917
            self,
            env: BuildEnvironment,
            fromdocname: str,
            builder: Builder,
            type: str,  # noqa: A002
            target: str,
            node: resolve_xref,
            contnode: Element,
        ) -> Element:
            # fixup some wrongly resolved mappings
            mapping = {
                "pathlib._local.Path": "pathlib.Path",
            }
            if target in mapping:
                target = node["reftarget"] = mapping[target]
            return super().resolve_xref(env, fromdocname, builder, type, target, node, contnode)

    app.add_domain(PatchedPythonDomain, override=True)



"""
Please keep this file Python 2.7 compatible.
See https://tox.readthedocs.io/en/rewrite/development.html#code-style-guide
"""

from __future__ import annotations

import os
import sys
import tarfile
from pathlib import Path
from textwrap import dedent
from zipfile import ZipFile

name = "demo_pkg_inline"
pkg_name = name.replace("_", "-")

version = "1.0.0"
dist_info = f"{name}-{version}.dist-info"
logic = f"{name}/__init__.py"
metadata_file = f"{dist_info}/METADATA"
wheel = f"{dist_info}/WHEEL"
record = f"{dist_info}/RECORD"
content = {
    logic: f"def do():\n    print('greetings from {name}')",
}
metadata = {
    metadata_file: f"""
        Metadata-Version: 2.1
        Name: {pkg_name}
        Version: {version}
        Summary: UNKNOWN
        Home-page: UNKNOWN
        Author: UNKNOWN
        Author-email: UNKNOWN
        License: UNKNOWN
        Platform: UNKNOWN

        UNKNOWN
       """,
    wheel: f"""
        Wheel-Version: 1.0
        Generator: {name}-{version}
        Root-Is-Purelib: true
        Tag: py{sys.version_info[0]}-none-any
       """,
    f"{dist_info}/top_level.txt": name,
    record: f"""
        {name}/__init__.py,,
        {dist_info}/METADATA,,
        {dist_info}/WHEEL,,
        {dist_info}/top_level.txt,,
        {dist_info}/RECORD,,
       """,
}


def build_wheel(
    wheel_directory: str,
    metadata_directory: str | None = None,
    config_settings: dict[str, str] | None = None,  # noqa: ARG001
) -> str:
    base_name = f"{name}-{version}-py{sys.version_info[0]}-none-any.whl"
    path = Path(wheel_directory) / base_name
    with ZipFile(str(path), "w") as zip_file_handler:
        for arc_name, data in content.items():
            zip_file_handler.writestr(arc_name, dedent(data).strip())
        if metadata_directory is not None:
            for sub_directory, _, filenames in os.walk(metadata_directory):
                for filename in filenames:
                    zip_file_handler.write(
                        str(Path(metadata_directory) / sub_directory / filename),
                        str(Path(sub_directory) / filename),
                    )
        else:
            for arc_name, data in metadata.items():
                zip_file_handler.writestr(arc_name, dedent(data).strip())
    print(f"created wheel {path}")  # noqa: T201
    return base_name


def get_requires_for_build_wheel(config_settings: dict[str, str] | None = None) -> list[str]:  # noqa: ARG001
    return []  # pragma: no cover # only executed in non-host pythons


def build_sdist(sdist_directory: str, config_settings: dict[str, str] | None = None) -> str:  # noqa: ARG001
    result = f"{name}-{version}.tar.gz"
    with tarfile.open(str(Path(sdist_directory) / result), "w:gz") as tar:
        root = Path(__file__).parent
        tar.add(str(root / "build.py"), "build.py")
        tar.add(str(root / "pyproject.toml"), "pyproject.toml")
    return result


def get_requires_for_build_sdist(config_settings: dict[str, str] | None = None) -> list[str]:  # noqa: ARG001
    return []  # pragma: no cover # only executed in non-host pythons


if "HAS_REQUIRES_EDITABLE" in os.environ:

    def get_requires_for_build_editable(config_settings: dict[str, str] | None = None) -> list[str]:  # noqa: ARG001
        return [1] if "REQUIRES_EDITABLE_BAD_RETURN" in os.environ else ["editables"]  # type: ignore[list-item]


if "HAS_PREPARE_EDITABLE" in os.environ:

    def prepare_metadata_for_build_editable(
        metadata_directory: str,
        config_settings: dict[str, str] | None = None,  # noqa: ARG001
    ) -> str:
        dest = Path(metadata_directory) / dist_info
        dest.mkdir(parents=True)
        for arc_name, data in metadata.items():
            (dest.parent / arc_name).write_text(dedent(data).strip())
        print(f"created metadata {dest}")  # noqa: T201
        if "PREPARE_EDITABLE_BAD" in os.environ:
            return 1  # type: ignore[return-value] # checking bad type on purpose
        return dist_info


def build_editable(
    wheel_directory: str,
    metadata_directory: str | None = None,
    config_settings: dict[str, str] | None = None,
) -> str:
    if "BUILD_EDITABLE_BAD" in os.environ:
        return 1  # type: ignore[return-value] # checking bad type on purpose
    return build_wheel(wheel_directory, metadata_directory, config_settings)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import pytest

import pyproject_api.__main__
from pyproject_api._frontend import EditableResult, SdistResult, WheelResult

if TYPE_CHECKING:
    import pytest_mock


@pytest.mark.parametrize(
    ("cli_args", "srcdir", "outdir", "hooks"),
    [
        (
            [],
            Path.cwd(),
            Path.cwd() / "dist",
            ["build_sdist", "build_wheel"],
        ),
        (
            ["src"],
            Path("src"),
            Path("src") / "dist",
            ["build_sdist", "build_wheel"],
        ),
        (
            ["-o", "out"],
            Path.cwd(),
            Path("out"),
            ["build_sdist", "build_wheel"],
        ),
        (
            ["-s"],
            Path.cwd(),
            Path.cwd() / "dist",
            ["build_sdist"],
        ),
        (
            ["-w"],
            Path.cwd(),
            Path.cwd() / "dist",
            ["build_wheel"],
        ),
        (
            ["-e"],
            Path.cwd(),
            Path.cwd() / "dist",
            ["build_editable"],
        ),
        (
            ["-s", "-w"],
            Path.cwd(),
            Path.cwd() / "dist",
            ["build_sdist", "build_wheel"],
        ),
    ],
)
def test_parse_args(
    mocker: pytest_mock.MockerFixture,
    capsys: pytest.CaptureFixture[str],
    cli_args: list[str],
    srcdir: Path,
    outdir: Path,
    hooks: list[str],
) -> None:
    subprocess_frontend = mocker.patch("pyproject_api.__main__.SubprocessFrontend", autospec=True)
    subprocess_frontend.create_args_from_folder.return_value = (srcdir, (), "foo.bar", "baz", (), True)
    subprocess_frontend.return_value.build_sdist.return_value = SdistResult(
        sdist=outdir / "foo.whl",
        out="sdist out",
        err="sdist err",
    )
    subprocess_frontend.return_value.build_wheel.return_value = WheelResult(
        wheel=outdir / "foo.whl",
        out="wheel out",
        err="wheel err",
    )
    subprocess_frontend.return_value.build_editable.return_value = EditableResult(
        wheel=outdir / "foo.whl",
        out="editable wheel out",
        err="editable wheel err",
    )

    pyproject_api.__main__.main(cli_args)

    subprocess_frontend.create_args_from_folder.assert_called_once_with(srcdir)
    captured = capsys.readouterr()

    if "build_sdist" in hooks:
        assert "Building sdist..." in captured.out
        subprocess_frontend.return_value.build_sdist.assert_called_once_with(outdir)
        assert "sdist out" in captured.out
        assert "sdist err" in captured.err

    if "build_wheel" in hooks:
        assert "Building wheel..." in captured.out
        subprocess_frontend.return_value.build_wheel.assert_called_once_with(outdir)
        assert "wheel out" in captured.out
        assert "wheel err" in captured.err

    if "build_editable" in hooks:
        assert "Building editable wheel..." in captured.out
        subprocess_frontend.return_value.build_editable.assert_called_once_with(outdir)
        assert "editable wheel out" in captured.out
        assert "editable wheel err" in captured.err



from __future__ import annotations

import json
import os
from typing import TYPE_CHECKING, Any

import pytest

from pyproject_api._backend import BackendProxy, read_line, run

if TYPE_CHECKING:
    from pathlib import Path

    import pytest_mock


def test_invalid_module(capsys: pytest.CaptureFixture[str]) -> None:
    with pytest.raises(ImportError):
        run([str(False), "an.invalid.module"])

    captured = capsys.readouterr()
    assert "failed to start backend" in captured.err


def test_invalid_request(mocker: pytest_mock.MockerFixture, capsys: pytest.CaptureFixture[str]) -> None:
    """Validate behavior when an invalid request is issued."""
    command = "invalid json"

    backend_proxy = mocker.MagicMock(spec=BackendProxy)
    backend_proxy.return_value = "dummy_result"
    backend_proxy.__str__.return_value = "FakeBackendProxy"
    mocker.patch("pyproject_api._backend.BackendProxy", return_value=backend_proxy)
    mocker.patch("pyproject_api._backend.read_line", return_value=bytearray(command, "utf-8"))

    ret = run([str(False), "a.dummy.module"])

    assert ret == 0
    captured = capsys.readouterr()
    assert "started backend " in captured.out
    assert "Backend: incorrect request to backend: " in captured.err


def test_exception(mocker: pytest_mock.MockerFixture, capsys: pytest.CaptureFixture[str], tmp_path: Path) -> None:
    """Ensure an exception in the backend is not bubbled up."""
    result = str(tmp_path / "result")
    command = json.dumps({"cmd": "dummy_command", "kwargs": {"foo": "bar"}, "result": result})

    backend_proxy = mocker.MagicMock(spec=BackendProxy)
    backend_proxy.side_effect = SystemExit(1)
    backend_proxy.__str__.return_value = "FakeBackendProxy"
    mocker.patch("pyproject_api._backend.BackendProxy", return_value=backend_proxy)
    mocker.patch("pyproject_api._backend.read_line", return_value=bytearray(command, "utf-8"))

    ret = run([str(False), "a.dummy.module"])

    # We still return 0 and write a result file. The exception should *not* bubble up
    assert ret == 0
    captured = capsys.readouterr()
    assert "started backend FakeBackendProxy" in captured.out
    assert "Backend: run command dummy_command with args {'foo': 'bar'}" in captured.out
    assert "Backend: Wrote response " in captured.out
    assert "SystemExit: 1" in captured.err


def test_valid_request(mocker: pytest_mock.MockerFixture, capsys: pytest.CaptureFixture[str], tmp_path: Path) -> None:
    """Validate the "success" path."""
    result = str(tmp_path / "result")
    command = json.dumps({"cmd": "dummy_command", "kwargs": {"foo": "bar"}, "result": result})

    backend_proxy = mocker.MagicMock(spec=BackendProxy)
    backend_proxy.return_value = "dummy-result"
    backend_proxy.__str__.return_value = "FakeBackendProxy"
    mocker.patch("pyproject_api._backend.BackendProxy", return_value=backend_proxy)
    mocker.patch("pyproject_api._backend.read_line", return_value=bytearray(command, "utf-8"))

    ret = run([str(False), "a.dummy.module"])

    assert ret == 0
    captured = capsys.readouterr()
    assert "started backend FakeBackendProxy" in captured.out
    assert "Backend: run command dummy_command with args {'foo': 'bar'}" in captured.out
    assert "Backend: Wrote response " in captured.out
    assert not captured.err


def test_reuse_process(mocker: pytest_mock.MockerFixture, capsys: pytest.CaptureFixture[str], tmp_path: Path) -> None:
    """Validate behavior when reusing the backend proxy process.

    There are a couple of things we'd like to check here:

    - Ensure we can actually reuse the process.
    - Ensure an exception in a call to the backend does not affect subsequent calls.
    - Ensure we can exit safely by calling the '_exit' command.
    """
    results = [
        str(tmp_path / "result_a"),
        str(tmp_path / "result_b"),
        str(tmp_path / "result_c"),
        str(tmp_path / "result_d"),
    ]
    commands = [
        json.dumps({"cmd": "dummy_command_a", "kwargs": {"foo": "bar"}, "result": results[0]}),
        json.dumps({"cmd": "dummy_command_b", "kwargs": {"baz": "qux"}, "result": results[1]}),
        json.dumps({"cmd": "dummy_command_c", "kwargs": {"win": "wow"}, "result": results[2]}),
        json.dumps({"cmd": "_exit", "kwargs": {}, "result": results[3]}),
    ]

    def fake_backend(name: str, *args: Any, **kwargs: Any) -> Any:  # noqa: ARG001
        if name == "dummy_command_b":
            raise SystemExit(2)

        return "dummy-result"

    backend_proxy = mocker.MagicMock(spec=BackendProxy)
    backend_proxy.side_effect = fake_backend
    backend_proxy.__str__.return_value = "FakeBackendProxy"
    mocker.patch("pyproject_api._backend.BackendProxy", return_value=backend_proxy)
    mocker.patch("pyproject_api._backend.read_line", side_effect=[bytearray(x, "utf-8") for x in commands])

    ret = run([str(True), "a.dummy.module"])

    # We still return 0 and write a result file. The exception should *not* bubble up and all commands should execute.
    # It is the responsibility of the caller to handle errors.
    assert ret == 0
    captured = capsys.readouterr()
    assert "started backend FakeBackendProxy" in captured.out
    assert "Backend: run command dummy_command_a with args {'foo': 'bar'}" in captured.out
    assert "Backend: run command dummy_command_b with args {'baz': 'qux'}" in captured.out
    assert "Backend: run command dummy_command_c with args {'win': 'wow'}" in captured.out
    assert "SystemExit: 2" in captured.err


def test_read_line_success() -> None:
    r, w = os.pipe()
    try:
        line_in = b"this is a line\r\n"
        os.write(w, line_in)
        line_out = read_line(fd=r)
        assert line_out == bytearray(b"this is a line")
    finally:
        os.close(r)
        os.close(w)


def test_read_line_eof_before_newline() -> None:
    r, w = os.pipe()
    try:
        line_in = b"this is a line"
        os.write(w, line_in)
        os.close(w)
        line_out = read_line(fd=r)
        assert line_out == bytearray(b"this is a line")
    finally:
        os.close(r)


def test_read_line_eof_at_the_beginning() -> None:
    r, w = os.pipe()
    try:
        os.close(w)
        with pytest.raises(EOFError):
            read_line(fd=r)
    finally:
        os.close(r)
